# A set of common helper functions usefull for iMX28 updaters

### Immediate execution

UPDDIR="$(dirname "${UPDFILE}")"
THISCODENAME="$(grep CODENAME /etc/updater/updater.conf | cut -d= -f2)"
FBSPLASH_SIZE="$(cut -f 2 /proc/navico_platform/disp_size | cut -c0)"

case "$THISCODENAME" in
    sylvester4g)
        FBSPLASH_SIZE="${FBSPLASH_SIZE}g"
        ;;
    *)
        ;;
esac

### Functions


nand_geometry() {
    local ERASESIZE=$(($(cat /sys/class/mtd/mtd0/erasesize) / 1024))
    local PAGESIZE=$(($(cat /sys/class/mtd/mtd0/writesize) / 1024))

    echo ${PAGESIZE}k${ERASESIZE}k
}

nand_page_size() {
    cat /sys/class/mtd/mtd0/writesize
}

nand_erase_size() {
    cat /sys/class/mtd/mtd0/erasesize
}

## Try and figure out the NAND size that is present
nand_size() {
    local SIZES=$(cat /proc/mtd | grep mtd | cut -d " " -f 2)
    local NANDSIZE=0

    # Sum the MTD partitions. This is a bit flawed, because there is no rule that mtd partitions
    # must use the whole device, or that the mtd partitions cannot overlap. Our mx51 platform
    # uses the whole device in non-overlapping partitions, so this will work.
    for size in $SIZES; do
        size=$((0x$size))
        NANDSIZE=$(($NANDSIZE + $size))
    done
    unset size

    echo $NANDSIZE
}

# Finds the ubi in format of eg ubi0_1
find_ubi_volume_num_from_name() {
    volume_name=$1

    # Scan the names of the ubi volumes, to check if this component is a volume image
    # Relies on UBI volumes not getting the same name as mtd partitions
    local UBIVOL=
    for syspath in /sys/class/ubi/ubi?_?; do
        if [ "$(cat ${syspath}/name)" = "${volume_name}" ]; then
            UBIVOL=$(echo ${syspath} | cut -d "/" -f 5)
            break
        fi
    done
    unset syspath
    echo $UBIVOL
}

get_upd_component() {
    # Pipe the named updater component to stdout.
    local name=$1

    if [ -s "$name" ]; then
        # If "name" refers to an existing file, use it.
        cat "$name"
    else
        # Otherwise assume "name" is in the updater tarball.
        tar -xOf "$UPDFILE" "${COMPONENT_PREFIX:-}${name}"
    fi
}

mtdpart() {
    grep -i \"$1\" /proc/mtd | sed 's/:.*//'
}


# Update a new rescue kernel on a legacy mtd layout.
update_rkernel() {
    local kernel_mtd_size=$(( 0x$(grep -i \"kernel\" /proc/mtd | cut -d' ' -f2) ))

    get_upd_component "$1" | xzcat > kernel.bin

    flash_eraseall /dev/$(mtdpart kernel)
    dd if=kernel.bin bs=$kernel_mtd_size count=1 | nandwrite -p /dev/$(mtdpart kernel)

    flash_eraseall /dev/$(mtdpart rkernel)
    dd if=kernel.bin bs=$kernel_mtd_size skip=1  | nandwrite -p /dev/$(mtdpart rkernel)
}

# Expects bootloader file name in $1, minimum bootloader version in $2
update_boot() {
    local version=$2
    local major_version=$(echo $version | cut -f1 -d .)
    local minor_version=$(echo $version | cut -f2 -d .)
    local update_required=0

    local current_major_ver=$(cut -f1  /proc/navico_platform/boot_ver)
    local current_minor_ver=$(cut -f2  /proc/navico_platform/boot_ver)

    if [ "$version" = "-1" ]; then
        update_required=1
    else
        if [ ${major_version} -gt ${current_major_ver} ]; then
            update_required=1
        elif [ ${major_version} -eq ${current_major_ver} ]; then
            if [ ${minor_version} -gt ${current_minor_ver} ]; then
                update_required=1
            else
                update_required=0
            fi
        else
            update_required=0
        fi
    fi

    if [ $update_required -eq 1 ]; then
        mkdir boottmp
        get_upd_component "$1" | xzcat | tar -x -C boottmp

        local envpad=boottmp/envpad-arm
        local env=boottmp/all.env
        local uboot_bin=boottmp/boot.bin
        local uboot_new=boottmp/boot.bin.new
        local kobs_ng=boottmp/kobs-ng

        if [ ! -x "$envpad" -o ! -f "$env" -o ! -s "$uboot_bin" -o ! -x "$kobs_ng" ]; then
            printf "Invalid bootloader component, aborting.\n" >&2
            return 1
        fi

        # If not forcing the CODENAME then preserve any existing one
        if [ -z "$CONVCODENAME" -a "${IGNORE_CONVCODENAME:-}" != "true" ]; then
            CONVCODENAME=$(sed -n 's/.*navico_sku\.codename=\([^ ]*\).*/\1/p' /proc/cmdline)
        fi

        if [ -n "${CONVCODENAME:-}" ]; then
            sed -i 's/navico_sku=${navico_sku}/& navico_sku.codename='"${CONVCODENAME}"'/' $env
        fi

        local env_offset=$( ${envpad} --info | sed 's/.*offset=\([0-9]*\).*/\1/' )

        ${envpad} --info

        (dd if=${uboot_bin} bs=${env_offset} conv=notrunc conv=sync count=1; cat ${env} | ${envpad} -) > $uboot_new

        local ubootmtd=$(mtdpart uboot)

        # If this boot uses the legacy mtd layout, update u-boot then reboot to use
        # the new mtd layout.
        if [ -n "$(mtdpart kernel)" ]; then
            # Write a new u-boot to the legacy partition id.
            flash_eraseall /dev/$ubootmtd
            nandwrite -p /dev/$ubootmtd $uboot_new
            reboot -f
        fi

        $kobs_ng init -v -w --chip_0_device_path=/dev/$ubootmtd $uboot_new

        rm -fR boottmp/
    fi
}

# expects mtd name in $1, file name in $2
update_mtd_image() {
    local mtd=$(mtdpart $1)
    if [ -n "${mtd}" ]; then
        flash_eraseall /dev/${mtd}
        get_upd_component "$2" | xzcat | nandwrite -p /dev/${mtd} -
    fi
}

check_updated_skel_files() {
    mount /mnt
    mount /mnt/home

    (
	cd /mnt/home
	USERS=$(find -type d -mindepth 1 -maxdepth 1)

	cd /mnt/etc/skel

	for user in $USERS; do
            for f in $(find -type f); do
		if ! cmp "$f" "/mnt/home/$user/$f"; then
                    echo "updating $f for $user"
                    cat "$f" > "/mnt/home/$user/$f"
		fi
            done
	done
    )

    sync
    umount /mnt/home
    umount /mnt
}

start_splash_screen() {
    sha1check fbsplash.tar.xz && tar -xOf "${UPDFILE}" fbsplash.tar.xz | xzcat | tar -x || true
    sha1check fbsplash-${THISCODENAME}.tar.xz && tar -xOf "${UPDFILE}" fbsplash-${THISCODENAME}.tar.xz | xzcat | tar -x || true
    if [ -e updating${FBSPLASH_SIZE}.ppm ]; then
        mkfifo fbsplash.fifo
        tar -xOf "${UPDFILE}" fbsplash.tar.xz | xzcat | tar -x || \
            tar -xOf "${UPDFILE}" fbsplash-${THISCODENAME}.tar.xz | xzcat | tar -x || \
            true
        fbsplash -s updating${FBSPLASH_SIZE}.ppm -i fbsplash${FBSPLASH_SIZE}.cfg -f fbsplash.fifo &
        sleep 2
    fi
}

display_progress () {
    if [ -e fbsplash.fifo ]; then
        echo $1 >> fbsplash.fifo
    fi
    return 0;
}

display_done_image() {
    if [ -e fbsplash.fifo ]; then
        echo exit >> fbsplash.fifo
        sleep 2
        rm fbsplash.fifo
    fi
    fbsplash -s updatefinished${FBSPLASH_SIZE}.ppm || true
    return 0;
}

display_error_image() {
    if [ -e fbsplash.fifo ]; then
        echo exit >> fbsplash.fifo
        sleep 2
        rm fbsplash.fifo
    fi
    fbsplash -s updateerror${FBSPLASH_SIZE}.ppm || true
    return 0;
}

check_repair_config_partition() {
    mount /mnt

    if ! mount -o ro /mnt/etc/NOS; then
        echo "Config parition not detected. Repairing..."

        mv /mnt/etc/NOS /mnt/etc/NOS.backup
        mkdir /mnt/etc/NOS
        mount -o rw /mnt/etc/NOS
        cp -a /mnt/etc/NOS.backup/* /mnt/etc/NOS/
        sync
        rm -fR /mnt/etc/NOS.backup
    fi

    umount /mnt/etc/NOS
    umount /mnt
}

reboot_on_exit() {
    echo 'reboot -f' > ~/.updater/bootaction
}

wait_for_media_reboot() {

    # Get the device number (in decimal) of updater
    MEDIA_DEV=$(stat -c '%d' "${UPDFILE}")
    # Generate a regex to use with /proc/partitions
    MEDIA_RE="^[[:space:]]*$(($MEDIA_DEV/256))[[:space:]]\+"       # major
    MEDIA_RE="${MEDIA_RE}$(($MEDIA_DEV%256))[[:space:]]\+"         # minor
    # Wait for mmcblk or USB block devices (e.g. Hobart)
    MEDIA_RE="${MEDIA_RE}[[:digit:]]\+[[:space:]]\+\(mmcblk\|sd\)" # blocks, name

    set +x  # squelch annoying infinite loop.
    # Omit -q(uiet) flag for first grep
    if grep "${MEDIA_RE}" /proc/partitions; then
        while grep -q "${MEDIA_RE}" /proc/partitions; do
            :
        done
    fi

    reboot_on_exit
}

showmsg() {
    local tty=/dev/tty1

    {
        local size_x=$(stty size|cut -d' ' -f2)
        local size_y=$(stty size|cut -d' ' -f1)
        local line_ofs=$((1+(size_y-${#})/2)) line=0
        clear
        for msg; do
            msg="\033[$((line_ofs+line));$(((size_x-${#msg})/2))H${msg}"
            [ $line -eq 0 ] && msg="\033[1;42m${msg}\033[0m"
            echo -e "$msg"
            line=$((line+1))
        done
    } <$tty >$tty
}
