#!/bin/sh
# gpg signed by build@mfd-gw-aklnz at @1698083288.
# Update script

# System handshake magic
UNAME=armv5tejl
CODENAME=charlieyanmar
CODENAMELIST=charlieyanmar
DATESTAMP=20230526210440

FAMILYID="$(cut -f2 /proc/navico_platform/family_id)"
SUBFAMILYID="$(cut -f2 /proc/navico_platform/sub_family_id)"

case "${FAMILYID}-${SUBFAMILYID}" in
*)
    MINBOOTVER=6.1
    ;;
esac

# This progress bar delay is just used as an approximate speed for moving the progress bar
# After the bar has passed 75%, it will slow to 1/2 this rate, and 1/4 this rate after 90%
PROGRESSBARDELAY=0.75

# Print out the updater version. This needed by the factory to validate the correct version is being
# programmed when doing automated CPU card programming
# NOTE: makeupd.sh will modify this line
echo UPDATEVERSION=YD42-1.1-67.5.14-35.4-4-gc23bad7acd

DEBUG=${DEBUG:-false}
if [ -e /dev/shm/debug ]; then
    DEBUG=true
fi
${DEBUG} && set -x

# If anything goes awry, exit early.
set -e

# We expect to be invoked from the /tmp directory.
# We expect to have working /proc and /sys directories, as well as a populated /dev

### This upgrade script expects the following environment variables to be present:
# UPDFILE -- a path to the '.upd' update file.
#    ex:  UPDFILE = /path/to/foo.upd
#    This file path is found during the discovery phase of the initramfs
#    and will be relative to the ramfs root.
if [ -z "${UPDFILE}" ]; then
    echo "Internal error: Bad environment.  UPDFILE was not set."
    exit 1;
fi

# HFP-1245: Insert template code from scripts/update_common.sh
find_component_sha1() {
    local ref_sha1="$(
    grep -A 100 "^###SHA1MAGIC###$" "${SHA1MAGIC_SRC:-${0}}" | \
    grep -o "[^ ]\{${SHA1SIZE:-40}\}  ${1}$"
    )"

    if [ $? -ne 0 -o -z "$ref_sha1" ]; then
    printf 'File "%s" not found in internal hash table.\n' "${1}" >&2
    return 1
    fi

    echo "$ref_sha1" | cut -f 1 -d ' '
}

sha1check() {
    local sha1cmd=${SHA1CMD:-sha1sum} external=0

    if [ "${1}" = "-external" ]; then
    external=1
    shift 1
    fi

    local file="${1}"
    local act_sha1 ref_sha1=$(find_component_sha1 "$file")

    if [ -z "$ref_sha1" ]; then
    # Error message printed by find_component_sha1()
    return 1
    fi

    [ "$DEBUG" = true ] && printf 'Stored SHA1 for "%s" is %s\n' "$file" "$ref_sha1" >&2

    local f_copy
    [ -d "${SHA1CHECK_SAVEDIR:-}" ] && f_copy="${SHA1CHECK_SAVEDIR}/${file}"

    if [ $external = 1 ]; then
    act_sha1="$(cat "$(dirname "${UPDFILE}")/${file}"            | tee ${f_copy+"$f_copy"} | $sha1cmd | cut -f 1 -d ' ')"
else
    act_sha1="$(tar -xOf "${UPDFILE}" ${COMPONENT_PREFIX:-}${file} | tee ${f_copy+"$f_copy"} | $sha1cmd | cut -f 1 -d ' ')"
    fi

    [ "$DEBUG" = true ] && printf 'Calculated SHA1 for "%s" is %s\n' "$file" "$act_sha1" >&2

    if [ "${ref_sha1}" != "${act_sha1}" ]; then
    printf 'ERROR: Update corrupt.  File "%s" failed checksum.\n' "$file"
    printf 'Please delete this update and replace it with a new copy.\n'
    return 1
else
    [ "$DEBUG" = true ] && printf 'DEBUG: successful sha1sum check: %s\n' "$file" >&2
    return 0
    fi
}

sha1check libupdate.sh || exit 1

tar -xOf "${UPDFILE}" "${COMPONENT_PREFIX:-}"libupdate.sh > libupdate.sh

. ./libupdate.sh


# Taken from zzz_imx:/usr/libexec/updater/selectupdate
get_last_update()
{
    local mnt=/mnt/home
    mkdir -p $mnt

    if mount -o ro $mnt ; then
        if [ -r $mnt/update-history.txt ]; then
	    # get datestamp:filename of last update, if successful
            # substitute 0 for missing datestamp (when NF==2)
	    awk -F$'\t' 'END {
                           if ($NF==0) {
                               printf "%s:%s\n",  (NF>=3 ? $2 : 0), $1
                           }
                         }' < $mnt/update-history.txt
        fi

        umount $mnt
    fi
    rmdir $mnt || true
}

# Taken from zzz_imx:/usr/libexec/updater/selectupdate: to_runupdate()
log_update_pending()
{
    local updatefile="$1"
    local datestamp="$2"

    local mnt=/mnt/home
    mkdir -p $mnt

    if mount $mnt ; then
        # mark as pending for now
        printf "%s\t%s\tP\n" "$(basename "${updatefile}")" "$datestamp" >> $mnt/update-history.txt
        sync
        umount $mnt
    fi
    rmdir $mnt || true
}

# Taken from zzz_imx:/sbin/update where the return code from runupdate is logged
log_update_result() {
    local result=$1
    local mnt=/mnt/home
    mkdir -p $mnt

    if mount $mnt ; then
        if [ -e $mnt/update-history.txt ]; then
	    sed -i "$ s/^\(.*\)\tP$/\1\t${result}/" $mnt/update-history.txt
	    sync
        fi
        umount $mnt
    else
        echo "Failed to log update result" >&2
    fi
    rmdir $mnt || true
}

case "$CODENAME" in
houdini)
    # if rescue kernel, install anyway (old behaviour)
    if grep -q "rescue" /proc/cmdline; then
        echo "Running latest updater - rescue kernel" >&2
        log_update_pending "$UPDFILE" "$DATESTAMP"
    else
        # datestamp:filename of last update, if successful
        last_update=$(get_last_update)

        # skip candidate unless name or date different from previous
        if [ "$last_update" != "${DATESTAMP}:$(basename "${UPDFILE}")" ]; then
            echo "Running latest updater" >&2
            log_update_pending "$UPDFILE" "$DATESTAMP"
        else
            echo "Skipping update to same version" >&2
            exit 0
        fi
    fi
    ;;
*)
    ;;
esac

start_splash_screen

# Background progress loop.
PROGRESS=0
(   set +x
    while [ ! -f /stop-progress-flag.$$ ] && \
        [ ${PROGRESS} -le 98 ];
    do
        PROGRESS=$((${PROGRESS}+1))
        display_progress ${PROGRESS}
        sleep ${PROGRESSBARDELAY}
        if [ ${PROGRESS} -ge 75 ]; then
             # 2x delay
             sleep ${PROGRESSBARDELAY}
        fi
        if [ ${PROGRESS} -ge 90 ]; then
             # 4x delay
             sleep ${PROGRESSBARDELAY}
             sleep ${PROGRESSBARDELAY}
        fi
    done
) &


###  Determine unit type

# What payloads are we carrying?
# Search our signature table.
COMPONENTLIST="$( \
    grep -A 42 -m 1 '^###SHA1MAGIC###$' "${0}" | \
    sed 's/[^ ]\{40\}  //; /###SHA1MAGIC###/d' | \
    sort | \
    uniq
)"

update_corrupt() {
    echo "ERROR:  This update is damaged.  Halted!"
    display_error_image
    while true; do
        sleep 42d
    done
}

# HFP-1542: Save boot and kernel components to tmpfs (so removing SD doesn't brick device)
CACHEDIR="/tmp/upd-cache-$$/"           # Note: Trailing / is required
mkdir -p "$CACHEDIR" || CACHEDIR=""

# Check all the components are valid before doing anything
for COMPONENT in $COMPONENTLIST; do
    case $COMPONENT in
    boot.bin|kernel.bin) SHA1CHECK_SAVEDIR="$CACHEDIR" ;;
    *)                   SHA1CHECK_SAVEDIR=""          ;;
    esac

    sha1check $COMPONENT || update_corrupt
done

cat /proc/mtd

# Replace the rescue kernel if this is a legacy mtd layout.
if [ -n "$(mtdpart kernel)" -a ! -z "$(echo ${COMPONENTLIST} | grep kernel.bin)" ]; then
    update_rkernel "${CACHEDIR}kernel.bin"
fi

# Install u-boot. This may reboot before returning so that we can use the most recent mtd layout. 
if [ ! -z "$(echo ${COMPONENTLIST} | grep boot.bin)" ]; then
    update_boot "${CACHEDIR}boot.bin" $MINBOOTVER
fi

# Repair any units that did not get config partition generated correctly in the factory
check_repair_config_partition

# Do the root file-system, so we get the tools on there for later use
if [ ! -z "$(echo ${COMPONENTLIST} | grep root.tar.xz)" ]; then
    # Delete and re-generate default file-system
    ubiupdatevol -t /dev/$(find_ubi_volume_num_from_name root)
    mount /mnt

    tar -xOf "${UPDFILE}" root.tar.xz | xzcat | tar -x -C /mnt

    umount /mnt
    sync
fi

mount /mnt
mount --bind /dev  /mnt/dev
mount --bind /sys  /mnt/sys
mount --bind /proc /mnt/proc

chroot /mnt sh -l -c "ldconfig"

[ -n "$CACHEDIR" ] && rm -fr "$CACHEDIR"

FBSPLASH_SIZE=$(cut -f 2 /proc/navico_platform/disp_size | cut -c0)

case "$CODENAME" in
sylvester4g)
    FBSPLASH_SIZE="${FBSPLASH_SIZE}g"
    ;;
*)
    ;;
esac

umount /mnt/proc
umount /mnt/sys
umount /mnt/dev
umount /mnt

check_updated_skel_files

echo "Update successful"

touch /stop-progress-flag.$$
sleep 1
display_progress 100
sleep 1

display_done_image

case "$CODENAME" in
houdini)
    reboot_on_exit
    log_update_result 0
    ;;
*)
    wait_for_media_reboot
    ;;
esac

# Done!
exit 0
exit 0
###SHA1MAGIC###
1ada4f086f15621b8c6d7f0b2543c86f83a2950c  libupdate.sh
fe7ae0bcbe2bb5d9495a28454fc62bca0d7bd34b  boot.bin
a59a96a1d4f63125de282e0ea3cc62fe2ee1f8a5  kernel.bin
5f0112e1279bff25c02a12653ab88f3d4bfcb6ca  root.tar.xz
8d7534084a0eae85ae9c2ab18701f9035614a7ac  fbsplash.tar.xz
