#!/bin/sh

# This script is used to determine if an updater is valid for the system upon
# which it is being run.
#
# COMMAND LINE
#
# One argument, a path to the file that needs testing
#
# RETURN CODE
#
# The return code is 0 if the script is successful, non-0 otherwise.
#
# OUTPUT
#
# The string 'true' is written to stdout, if the updater is valid.
# The string 'false: {reason}' is written to stdout, if the updater is not valid.
# The string 'error: {reason}' is written to stderr, if the script failed.
#
# NOTE: This code must be compatible with source/k/initrd-updater/*

set -u

source /lib/functions/navico_env.sh
source /lib/functions/upd_codename.sh
source /lib/functions/upd_datestamp.sh

verify() {
    if [ -z "$1" ]; then
        echo "error: missing argument" 1>&2
        return 1

    elif [ ! -r "$1" ]; then
        echo "error: unreadable file: '$1'" 1>&2
        return 1

    else
        local upd_tags
        navico_env_init

        if [ -z "$CODENAME" ]; then
            echo 'error: missing host codename' 1>&2
            return 1
        fi

        if ! upd_tags=$(
            tar -xOf "$1" update.sh 2>/dev/null |
            head -n50 |
            cut -f1 |
            cut -f1 -d'#' |
            grep -o '^\(CODENAME\(LIST\)\?\|DATESTAMP\|UPD_PRIORITY\|UPD_FLAGS\|\(echo \)\?UPDATEVERSION\)=[^ ]\+'
        ); then
            echo 'error: not an updater' 1>&2
            return 1
        fi

        init_codename_re

        if ! printf '%s' "$upd_tags" | grep -q "$CODENAME_RE"; then
            echo 'false: codename mismatch'
            return 0
        fi

        local datestamp=$(printf '%s' "$upd_tags" | grep 'DATESTAMP=' | cut -f2- -d'=')
        local upd_flags=$(printf '%s' "$upd_tags" | grep 'UPD_FLAGS=' | cut -f2- -d'=')

        if [ -z "$upd_flags" ] && printf '%s' "$upd_tags" | grep -q 'UPDATEVERSION='; then
            upd_flags='A'  # Assume "app/platform" updater
        fi

        # HFP-1004: Do not accept app/platform updaters earlier than factory DATESTAMP.
        if [ "$upd_flags" != "${upd_flags/A/}" ] && ! check_factory_datestamp "$datestamp"; then
            echo 'false: too old'
            return 0
        fi

        echo 'true'
    fi
}

verify "${1:-}"
exit $?
