#!/bin/sh

HDMI_DEV=/sys/devices/soc0/soc/20e0000.hdmi_video
EDID=$HDMI_DEV/edid
SAFE_MODE=640x480-60
FB=
SYSFB=

MODE_PER_MONITOR=0
MODES=/home/monitors

mode_file() {
    local monitor="$1"

    if [ "$MODE_PER_MONITOR" = "1" ]; then
	# store per-monitor settings
	echo "$MODES/$monitor"
    else
	# store one common setting
	echo "$MODES/hdmi_mode.txt"
    fi
}

manufacturer_id() {
    # nb we should never see digits
    local char="0ABCDEFGHIJKLMNOPQRSTUVWXYZ12345" # 0-31
    local edid=$1
    local m1 m2 m3
    if [ -e $edid ] ; then
        head -n1 $edid | cut -d' ' -f 9,10 | if read a b ; then
            m1="${char:$(((a>>2)&31)):1}"
            m2="${char:$((((a&3)<<3)|((b>>5)&7))):1}"
            m3="${char:$((b&31)):1}"
            echo ${m1}${m2}${m3}
        fi
    fi
}

# 16 bit little endian
# offset 10-11
product_code () {
    local edid=$1
    if [ -e $edid ] ; then
 	head -n1 $edid | cut -d' ' -f 11,12 | if read a b ; then
 	    code=$(((b*256)+a))
 	    echo ${code}
	fi
    fi
}

monitor_id() {
    local edid=$1
    echo $(manufacturer_id $edid)_$(product_code $edid)
}

# find the FB device that corresponds to mxc_hdmi
get_hdmi_fb() {
    local fb_name="$(cat ${HDMI_DEV}/fb_name)"

    for fb in /sys/class/graphics/fb* ; do
        if [ "$(cat ${fb}/name)" = "$fb_name" ]; then
            echo $fb | sed "s/.*\(fb[0-9]\+\)/\1/"
	    break;
        fi
    done
}

# D:1024x768p-60 => D 1024 768 60
# 1024x768p-60 => U 1024 768 60
parse_mode() {
    echo "$1" | sed -n \
	-e 's/^\([A-Z]\):\([0-9]\+\)x\([0-9]\+\)[a-z]\?-\([0-9]\+\)$/\1 \2 \3 \4/gp' \
	-e 's/^\([0-9]\+\)x\([0-9]\+\)[a-z]\?-\([0-9]\+\)$/U \1 \2 \3/gp'
}

# F:HxV-R
canonical_mode() {
    parse_mode "$1" | if read f h v r ; then
	echo "${f}:${h}x${v}p-${r}"
    fi
}

get_modelist() {
    cat $SYSFB/modes | tr DSVU abce | sort | tr abce DSVU | while read mode ; do
	canonical_mode "$mode"
    done
}

clear_fb() {
    dd if=/dev/zero of=/dev/$FB > /dev/null 2>&1
}

find_matching_mode() {
    local h=$1
    local v=$2
    local r=$3

    if [ -n "$r"  ]; then
	# resolution and refresh rate
	get_modelist | grep -e "^[A-Z]:${h}x${v}p-${r}$" | head -n1
    else
	# just resolution
	get_modelist | grep -e "^[A-Z]:${h}x${v}p-" | head -n1
    fi
}

get_preferred_mode() {
    get_modelist | grep "D:" | head -n1
}

get_active_mode() {
    mode=$(cat $SYSFB/mode)

    if [ -n "$mode" ]; then
	echo "$(canonical_mode "$mode")"
    else
	# if we used fbset to set mode, then $SYSFB/mode may be empty
	# se we use fbset to grab mode string
	echo "U:$(fbset -fb /dev/$FB | awk '/^mode/ { printf "%s\n", $2 }' | tr -d \")"
    fi
}

get_saved_mode() {
    local file="$(mode_file "$1")"
    [ -e "$file" ] && canonical_mode "$(cat "$file")"
}

save_mode() {
    local mode="$1"
    local file="$(mode_file "$2")"
    canonical_mode "$mode" > "$file"
}

clear_mode() {
    local file="$(mode_file "$1")"
    [ -e "$file" ] && rm -f "$file"
}

clear_all() {
    [ -d "$MODES" ] && rm -rf "$MODES"
    [ -e /home/hdmi_mode.txt ] && rm -f /home/hdmi_mode.txt
}

set_mode() {
    local mode="$1"
    local monitor="$2"
    [ -n "$mode" ] && save_mode "$mode" "$monitor"
}

get_added_mode() {
    local prev="$1"
    comm -13 "$prev" $SYSFB/modes
}

# apply_mode safe: force safe mode
# apply_mode <mode>: apply given mode
# apply_mode (no arguments):
#  - apply saved mode if any
#  - or apply preferred mode
apply_mode() {
    local mode
    local found

    # safe, user, or preferred mode
    if [ "$1" = "safe" ]; then
	mode="$SAFE_MODE"
    elif [ -n "$1" ]; then
	mode="$1"
    else
	mode="$(get_saved_mode "$ID")"

	if [ -z "$mode" ]; then
	    mode="$(get_preferred_mode)"
	fi
    fi

    # apply it to the framebuffer
    parse_mode "$mode" | if read f h v refresh ; then
	local m=$(find_matching_mode $h $v $refresh)
	local current_mode="$(cat "$SYSFB/mode")"
	local current_res="$(parse_mode "$current_mode" | cut -d' ' -f2,3)"
	local new_res="$h $v"

	if [ -n "$m" ]; then
	    if [ "$m" != "$current_mode" ]; then
		[ "$new_res" != "$current_res" ] && clear_fb
		echo "$m" > $SYSFB/mode
	    fi
	    # else: no change
	else
	    # new mode not in modelist
	    local prev="/tmp/${FB}_modes_$$.txt"
	    cat $SYSFB/modes > "$prev"

	    [ "$new_res" != "$current_res" ] && clear_fb

	    # note: gtf rounds up, so e.g. 1366->1368, is that a problem? hope not
	    # echo "Creating new mode" >&2
	    local depth=$(cat "$SYSFB/bits_per_pixel")
	    local name=baz
            gtf $h $v $refresh -f | awk '
                                    /^mode/ { printf "mode \"'$name'\"\n" }
                                    /geometry/ { printf "%s %s %s %s %s '$depth'\n", $1, $2, $3, $4, $5 }
                                    /timings|vsync|hsync|endmode/  { printf "%s\n", $0 }' | fbset -db - -fb /dev/$FB $name

	    N="$(get_added_mode "$prev")"
	    rm -f "$prev" > /dev/null 2>&1

	    if [ -n "$N" ]; then
		# echo "New mode added: $N" >&2
		echo "$N" > $SYSFB/mode
	    fi
	fi
    fi
}

if [ ! -d "$MODES" ]; then
    # this way, only "apply" needs sudo
    mkdir -m 775 -p "$MODES"
    chgrp users "$MODES"
fi

cable_state=$(cat ${HDMI_DEV}/cable_state)

if [ "$cable_state" = "plugin" ]; then
    FB=$(get_hdmi_fb)
    SYSFB=/sys/class/graphics/$FB
    ID=$(monitor_id "$EDID")
else
    echo "No HDMI monitor detected"
    exit 1
fi

case "$1" in
    modes)
	get_modelist
	;;
    preferred)
	get_preferred_mode
	;;
    saved)
	get_saved_mode "$ID"
	;;
    active)
	get_active_mode
	;;
    set)
	set_mode "$2" "$ID"
	;;
    clear)
	clear_mode "$ID"
	;;
    safe)
	# apply it but don't save it
	apply_mode "safe"
	;;
    apply)
	apply_mode "$2"
	;;
    id)
	echo "$(monitor_id ${2:-$EDID})"
	;;
    ""|status)
	local saved=$(get_saved_mode "$ID")
	echo "Monitor ID:            $ID"
	echo "Saved user mode:       $saved"
	echo "EDID Preferred mode:   $(get_preferred_mode)"
	echo "Currently active mode: $(get_active_mode)"
	echo "Mode timings:          $(fbset -fb /dev/$FB | awk '/timings/ { printf "%s %s %s %s %s %s %s\n", $2, $3, $4, $5, $6, $7, $8}')"
	echo "Mode pixel clock:      $(fbset -fb /dev/$FB | awk '/MHz/ { printf "%s MHz\n", $3}')"
	echo
	echo "Mode list for $FB:"
	get_modelist
	echo
	echo "'help' for more options"
	;;
    *)
	echo "Usage: $(basename "$0") modes | preferred | active | saved | set <mode> | clear | apply | safe | id | status | help"
	;;
esac
